#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <an.h>
#include "head.h"
#define NOISE	.01

double fuv( u, v, gamma, a )
	double u, v, gamma, a ;
{
	double tmp ;

	tmp = (a-u+u*u*v)*gamma ;

	return tmp ;
}

double guv( u, v, gamma, b )
	double u, v, gamma, b ;
{
	double tmp ;

	tmp = (b-u*u*v)*gamma ;

	return tmp ;
}

int set_initial_state( ps, ou, nu, ov, nv )
	PARAMS ps ;
	double **ou, **ov, **nu, **nv ;
{
	int x, y ;
	double u0, v0, hrm ;

	u0 = ps.a + ps.b ;
	v0 = ps.b/pow(u0,2.) ;
	hrm = (double)RAND_MAX/2. ;
	printf( " u0=%f, v0=%f, hrm=%f\n", u0, v0, hrm ) ;

	srand(0) ;
	for ( y=0; y<ps.ly; ++y )
	for ( x=0; x<ps.lx; ++x ) {
		ou[y][x] = nu[y][x] = u0 + NOISE*(rand()-hrm)/hrm ;
		ov[y][x] = nv[y][x] = v0 + NOISE*(rand()-hrm)/hrm ;
	}

	return 0 ;
}

int calc( ps, ou, nu, ov, nv, id, var )
	PARAMS ps ;
	double **ou, **nu, **ov, **nv, **id, **var ;
{
	int t, x, y ;
	double **d, dudt, dvdt, max_dudt, max_dvdt ;
	FILE *fpu, *fpv ;
	char str[STR_LEN] ;
	int put_msg(), get_image(), write_img(), sub() ;

	d=(double **)an_malloc_2D(ps.lx,ps.ly,sizeof**d) ;
	fpu=fopen(strcat(strcpy(str,ps.result_file),".u"),"wb") ;
	fpv=fopen(strcat(strcpy(str,ps.result_file),".v"),"wb") ;

	if ( d==NULL || fpu==NULL || fpv==NULL ) {
		free( d ) ;
		fclose( fpu ) ;
		fclose( fpv ) ;
		return -1 ;
	} else {
		put_msg( ps ) ;
		get_image( ps, id, var ) ;
		set_initial_state( ps, ou, nu, ov, nv ) ;
		write_img( ps, ou, ov, fpu, fpv ) ;
	}

	for ( t=1; t<ps.dt*ps.lt; ++t ) {
		if ( t%ps.tintv==0 ) {
			printf( "t=%f, u=%e, v=%e\n", (double)t/ps.dt,
				ou[ps.ly/2][ps.lx/2], ov[ps.ly/2][ps.lx/2] ) ;
			write_img( ps, ou, ov, fpu, fpv ) ;
		}

		sub( ps, ou, nu, ov, nv, d, id, var ) ;

		if ( t%ps.tintv==0 ) {
			max_dudt = max_dvdt = 0. ;
			for ( y=0; y<ps.ly; ++y )
			for ( x=0; x<ps.lx; ++x ) {
				dudt = fabs( ou[y][x]-nu[y][x] ) ;
				dvdt = fabs( ov[y][x]-nv[y][x] ) ;
				if ( dudt > max_dudt ) max_dudt = dudt ;
				if ( dvdt > max_dvdt ) max_dvdt = dvdt ;
			}
			printf( "max_dudt: %e	%e\n",
				(double)t/ps.dt, max_dudt*ps.dt ) ;
			printf( "max_dvdt: %e	%e\n",
				(double)t/ps.dt, max_dvdt*ps.dt ) ;
			fflush( stdout ) ;
		}

		for ( y=0; y<ps.ly; ++y )
		for ( x=0; x<ps.lx; ++x ) {
			ou[y][x] = nu[y][x] ;
			ov[y][x] = nv[y][x] ;
		}
	}

	write_img( ps, ou, ov, fpu, fpv ) ;

	free( d ) ;
	fclose( fpu ) ;
	fclose( fpv ) ;

	return 0 ;
}

int write_img( ps, ou, ov, fpu, fpv )
	PARAMS ps ;
	double **ou, **ov ;
	FILE *fpu, *fpv ;
{
	int y ;

	for ( y=0; y<ps.ly; ++y ) {
		fwrite( ou[y], sizeof **ou, ps.lx, fpu ) ;
		fwrite( ov[y], sizeof **ov, ps.lx, fpv ) ;
	}

	fflush( fpu ) ;

	return 0 ;
}

int put_msg( ps )
	PARAMS ps ;
{
	printf( "PARAMS: \n parameter file; %s, result file; %s \n",
			ps.para_file, ps.result_file ) ;
	printf( " dx=%d, dy=%d, dt=%d\n", ps.dx, ps.dy, ps.dt ) ;
	printf( " lx=%d, ly=%d, lt=%d\n", ps.lx, ps.ly, ps.lt ) ;
	printf( " Du=%e, Dv=%e\n a=%e, b=%e, gamma=%e\n var_min=%e, var_max=%e\n",
			ps.Du, ps.Dv,
			ps.a, ps.b, ps.gamma,
			ps.var_min, ps.var_max ) ;
	printf( " dt/dx^2 = %e, dt/dy^2=%e\n",
			(double)(ps.dx*ps.dx)/(double)ps.dt,
			(double)(ps.dy*ps.dy)/(double)ps.dt ) ;
	printf( " Loop=%d, Delta=%e, Ratio of C-N=%f\n",
			ps.loop, ps.delta, ps.r ) ;
	printf( " Interval: t=%d\n", ps.tintv ) ;
	printf( " File name of image: %s\n", ps.fnu ) ;
	printf( " norm=%f\n", ps.norm ) ;
	printf( " NOISE=%f\n", NOISE ) ;

	fflush( stdout ) ;

	return 0 ;
}

int mk_cu( ps, ou, ov, d, var )
	PARAMS ps ;
	double **ou, **ov, **d, **var ;
{
	int y, x, lx, ly ;
	int xl, xh, yl, yh ;
	double a, b, c, e ;

	a = 1. - 2. * ( 1. - ps.r ) * ps.Cux - 2. * ( 1. - ps.r ) * ps.Cuy ;
	b = ( 1. - ps.r ) * ps.Cux ;
	c = ( 1. - ps.r ) * ps.Cuy ;
	e = (double)ps.dt ;

	lx = ps.lx ;
	ly = ps.ly ;

	d[0][0] = a*ou[0][0]
		+ b*(ou[0][1]+ou[0][0])
		+ c*(ou[1][0]+ou[0][0])
		+ fuv( ou[0][0], ov[0][0], ps.gamma, ps.a ) / e ;

	d[0][lx-1] = a*ou[0][lx-1]
		+ b*(ou[0][lx-1]+ou[0][lx-2])
		+ c*(ou[1][lx-1]+ou[0][lx-1])
		+ fuv( ou[0][lx-1], ov[0][lx-1], ps.gamma, ps.a ) / e ;

	d[ly-1][0] = a*ou[ly-1][0]
		+ b*(ou[ly-1][1]+ou[ly-1][0])
		+ c*(ou[ly-1][0]+ou[ly-2][0])
		+ fuv( ou[ly-1][0], ov[ly-1][0], ps.gamma, ps.a ) / e ;

	d[ly-1][lx-1] = a*ou[ly-1][lx-1]
		+ b*(ou[ly-1][lx-1]+ou[ly-1][lx-2])
		+ c*(ou[ly-1][lx-1]+ou[ly-2][lx-1])
		+ fuv( ou[ly-1][lx-1], ov[ly-1][lx-1], ps.gamma, ps.a ) / e ;

	for ( x=1; x<lx-1; ++x ) {
		d[0][x] = a*ou[0][x]
			+ b*(ou[0][x+1]+ou[0][x-1])
			+ c*(ou[1][x]+ou[0][x])
			+ fuv( ou[0][x], ov[0][x], ps.gamma, ps.a ) / e ;

		d[ly-1][x] = a*ou[ly-1][x]
			+ b*(ou[ly-1][x+1]+ou[ly-1][x-1])
			+ c*(ou[ly-1][x]+ou[ly-2][x])
			+ fuv( ou[ly-1][x], ov[ly-1][x], ps.gamma, ps.a ) / e ;
	}

	for ( y=1; y<ly-1; ++y ) {
		yl=y-1 ;
		yh=y+1 ;
		d[y][0] = a*ou[y][0]
			+ b*(ou[y][1]+ou[y][0])
			+ c*(ou[yh][0]+ou[yl][0])
			+ fuv( ou[y][0], ov[y][0], ps.gamma, ps.a ) / e ;
		for ( x=1; x<lx-1; ++x ) {
			xl=x-1 ;
			xh=x+1 ;
			d[y][x] = a*ou[y][x]
				+ b*(ou[y][xh]+ou[y][xl])
				+ c*(ou[yh][x]+ou[yl][x])
				+ fuv( ou[y][x], ov[y][x],
					ps.gamma, ps.a ) / e ;
		}
		d[y][lx-1] = a*ou[y][lx-1]
			+ b*(ou[y][lx-1]+ou[y][lx-2])
			+ c*(ou[yh][lx-1]+ou[yl][lx-1])
			+ fuv( ou[y][lx-1], ov[y][lx-1], ps.gamma, ps.a ) / e ;
	}

	return 0 ;
}

int mk_cv( ps, ou, ov, d, var )
	PARAMS ps ;
	double **ou, **ov, **d, **var ;
{
	int x, y, lx, ly ;
	int xl, xh, yl, yh ;
	double a, b, c, e ;

	a = 1. - 2.*(1.-ps.r)*ps.Cvx - 2.*(1.-ps.r)*ps.Cvy ;
	b = ( 1. - ps.r ) * ps.Cvx ;
	c = ( 1. - ps.r ) * ps.Cvy ;
	e = (double)ps.dt ;

	lx = ps.lx ;
	ly = ps.ly ;

	d[0][0] = a*ov[0][0]
		+ b*(ov[0][1]+ov[0][0])
		+ c*(ov[1][0]+ov[0][0])
		+ guv( ou[0][0], ov[0][0], ps.gamma, var[0][0] ) / e ;

	d[0][lx-1] = a*ov[0][lx-1]
		+ b*(ov[0][lx-1]+ov[0][lx-2])
		+ c*(ov[1][lx-1]+ov[0][lx-1])
		+ guv( ou[0][lx-1], ov[0][lx-1], ps.gamma, var[0][lx-1] ) / e ;

	d[ly-1][0] = a*ov[ly-1][0]
		+ b*(ov[ly-1][1]+ov[ly-1][0])
		+ c*(ov[ly-1][0]+ov[ly-2][0])
		+ guv( ou[ly-1][0], ov[ly-1][0], ps.gamma, var[ly-1][0] ) / e ;

	d[ly-1][lx-1] = a*ov[ly-1][lx-1]
		+ b*(ov[ly-1][lx-1]+ov[ly-1][lx-2])
		+ c*(ov[ly-1][lx-1]+ov[ly-2][lx-1])
		+ guv( ou[ly-1][lx-1], ov[ly-1][lx-1],
			ps.gamma, var[ly-1][lx-1] ) / e ;

	for ( x=1; x<lx-1; ++x ) {
		d[0][x] = a*ov[0][x]
			+ b*(ov[0][x+1]+ov[0][x-1])
			+ c*(ov[1][x]+ov[0][x])
			+ guv( ou[0][x], ov[0][x], ps.gamma, var[0][x] ) / e ;
		d[ly-1][x] = a*ov[ly-1][x]
			+ b*(ov[ly-1][x+1]+ov[ly-1][x-1])
			+ c*(ov[ly-1][x]+ov[ly-2][x])
			+ guv( ou[ly-1][x], ov[ly-1][x],
				ps.gamma, var[ly-1][x] ) / e ;
	}

	for ( y=1; y<ly-1; ++y ) {
		yl=y-1 ;
		yh=y+1 ;
		d[y][0] = a*ov[y][0]
			+ b*(ov[y][1]+ov[y][0])
			+ c*(ov[yh][0]+ov[yl][0])
			+ guv( ou[y][0], ov[y][0], ps.gamma, var[y][0] ) / e ;
		for ( x=1; x<lx-1; ++x ) {
			xl=x-1 ;
			xh=x+1 ;
			d[y][x] = a*ov[y][x]
				+ b*(ov[y][xh]+ov[y][xl])
				+ c*(ov[yh][x]+ov[yl][x])
				+ guv( ou[y][x], ov[y][x],
					ps.gamma, var[y][x] ) / e ;
		}
		d[y][lx-1] = a*ov[y][lx-1]
			+ b*(ov[y][lx-1]+ov[y][lx-2])
			+ c*(ov[yh][lx-1]+ov[yl][lx-1])
			+ guv( ou[y][lx-1], ov[y][lx-1],
				ps.gamma, var[y][lx-1] ) / e ;
	}

	return 0 ;
}

int solve( ps, n, o, d, p, q, r, lx, ly )
	PARAMS ps ;
	double **n, **o, **d ;
	double p, q, r ;
	int lx, ly ;
{
	int l, x, y, xl, xh, yl, yh ;
	double sum, tmp ;

	for ( l=0; l<ps.loop; ++l ) {
		sum = .0 ;
		tmp = ( q*(n[0][1]+n[0][0]) + r*(n[1][0]+n[0][0])
				+ d[0][0] ) / p ;
		sum += (n[0][0]-tmp)*(n[0][0]-tmp) ;
		n[0][0] = tmp ;
		for ( x=1; x<lx-1; ++x ) {
			tmp = ( q*(n[0][x+1]+n[0][x-1])
				+ r*(n[1][x]+n[0][x]) + d[0][x] ) / p ;
			sum += (n[0][x]-tmp)*(n[0][x]-tmp) ;
			n[0][x] = tmp ;
		}
		tmp = ( q*(n[0][lx-1]+n[0][lx-2])
				+ r*(n[1][lx-1]+n[0][lx-1]) + d[0][lx-1] ) /p;
		sum += (n[0][lx-1]-tmp)*(n[0][lx-1]-tmp) ;
		n[0][lx-1] = tmp ;

		for ( y=1; y<ly-1; ++y ) {
			yl=y-1;
			yh=y+1;
			tmp = ( q*(n[y][1]+n[y][0])
				+ r*(n[yh][0]+n[yl][0]) + d[y][0] ) / p ;
			sum += (n[y][0]-tmp)*(n[y][0]-tmp) ;
			n[y][0] = tmp ;
			for ( x=1; x<lx-1; ++x ) {
				xl=x-1;
				xh=x+1;
				tmp = ( q*(n[y][xh]+n[y][xl])
					+ r*(n[yh][x]+n[yl][x])
					+ d[y][x] ) / p ;
				sum += (n[y][x]-tmp)*(n[y][x]-tmp) ;
				n[y][x] = tmp ;
			}
			tmp = ( q*(n[y][lx-1]+n[y][lx-2])
				+ r*(n[yh][lx-1]+n[yl][lx-1]) + d[y][lx-1])/p;
			sum += (n[y][lx-1]-tmp)*(n[y][lx-1]-tmp) ;
			n[y][lx-1] = tmp ;
		}
		tmp = ( q*(n[ly-1][1]+n[ly-1][0])
			+ r*(n[ly-1][0]+n[ly-2][0]) + d[ly-1][0] ) / p ;
		sum += (n[ly-1][0]-tmp)*(n[ly-1][0]-tmp) ;
		n[ly-1][0] = tmp ;
		for ( x=1; x<lx-1; ++x ) {
			tmp = ( q*(n[ly-1][x+1]+n[ly-1][x-1])
				+ r*(n[ly-1][x]+n[ly-2][x]) + d[ly-1][x] )/p ;
			sum += (n[ly-1][x]-tmp)*(n[ly-1][x]-tmp) ;
			n[ly-1][x] = tmp ;
		}
		tmp = ( q*(n[ly-1][lx-1]+n[ly-1][lx-2])
				+ r*(n[ly-1][lx-1]+n[ly-2][lx-1])
				+ d[ly-1][lx-1] ) / p ;
		sum += (n[ly-1][lx-1]-tmp)*(n[ly-1][lx-1]-tmp) ;
		n[ly-1][lx-1] = tmp ;

		if ( sum<ps.delta ) break ;
	}

	return 0 ;
}

int sub( ps, ou, nu, ov, nv, d, id, var )
	PARAMS ps ;
	double **ou, **nu, **ov, **nv, **d, **id, **var ;
{
	double p, q, r ;

	mk_cu( ps, ou, ov, d, var ) ;
	p = 1.+2.*ps.r*(ps.Cux+ps.Cuy) ;
	q = ps.Cux*ps.r ;
	r = ps.Cuy*ps.r ;
	solve( ps, nu, ou, d, p, q, r, ps.lx, ps.ly ) ;

	mk_cv( ps, ou, ov, d, var ) ;
	p = 1.+2.*ps.r*(ps.Cvx+ps.Cvy) ;
	q = ps.Cvx*ps.r ;
	r = ps.Cvy*ps.r ;
	solve( ps, nv, ov, d, p, q, r, ps.lx, ps.ly ) ;

	return 0 ;
}

int get_image( ps, image_data, var )
	PARAMS ps ;
	double **image_data, **var ;
{
	int x, y ;
	FILE *fp ;
	unsigned char *d ;
	double tmp ;

	if ( ( d = (unsigned char *)malloc( ps.lx * sizeof *d ) ) == NULL )
		return -1 ;
	else if ( ( fp = fopen( ps.fnu, "rb" ) ) == NULL )
		return -1 ;
	else {
		for ( y=0; y<ps.ly; ++y ) {
			fread( d, sizeof *d, ps.lx, fp ) ;
			for ( x=0; x<ps.lx; ++x )
				image_data[y][x] = (double)d[x]/ps.norm ;
		}

		fclose( fp ) ;
		free( d ) ;

		for ( y=0; y<ps.ly; ++y )
		for ( x=0; x<ps.lx; ++x ) {
			tmp = image_data[y][x] ;
			var[y][x] = ps.var_min + (ps.var_max-ps.var_min)*tmp ;
		}

		return 0 ;
	}
}

