#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <an.h>
#include "head.h"

#define COMMAND_NAME "sj"

int chk_para( fn, pps )
	char **fn ;
	PARAMS *pps ;
{
	FILE *fp ;

	(*pps).para_file = fn[1] ;
	(*pps).result_file = fn[2] ;

	if ( (fp=fopen((*pps).para_file,"r"))==(FILE *)NULL )
		return -1 ;
	else {
		char str[STR_LEN] ;
		int i ;

		(*pps).dx	= atoi( fgets( str, STR_LEN, fp ) ) ;
		(*pps).dy	= atoi( fgets( str, STR_LEN, fp ) ) ;
		(*pps).dt	= atoi( fgets( str, STR_LEN, fp ) ) ;

		(*pps).lx	= atoi( fgets( str, STR_LEN, fp ) ) ;
		(*pps).ly	= atoi( fgets( str, STR_LEN, fp ) ) ;
		(*pps).lt	= atoi( fgets( str, STR_LEN, fp ) ) ;

		(*pps).loop	= atoi( fgets( str, STR_LEN, fp ) ) ;
		(*pps).tintv	= atoi( fgets( str, STR_LEN, fp ) ) ;
		(*pps).r	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).delta	= atof( fgets( str, STR_LEN, fp ) ) ;

		fgets( (*pps).fnu, STR_LEN, fp ) ;
		for ( i=0; i<STR_LEN; ++i )
			if ( ((*pps).fnu)[i]=='\n' ) {
				((*pps).fnu)[i] = (char)NULL ;
				break ;
			}

		(*pps).Du	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).Dv	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).a	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).b	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).gamma	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).var_min	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).var_max	= atof( fgets( str, STR_LEN, fp ) ) ;
		(*pps).norm	= atof( fgets( str, STR_LEN, fp ) ) ;

		(*pps).Cux	= (*pps).Du *
					(double)(*pps).dx * (double)(*pps).dx
						 / (double)(*pps).dt ;
		(*pps).Cuy	= (*pps).Du *
					(double)(*pps).dy * (double)(*pps).dy
						 / (double)(*pps).dt ;
		(*pps).Cvx	= (*pps).Dv *
					(double)(*pps).dx * (double)(*pps).dx
						 / (double)(*pps).dt ;
		(*pps).Cvy	= (*pps).Dv *
					(double)(*pps).dy * (double)(*pps).dy
						 / (double)(*pps).dt ;

		fclose( fp ) ;
		return 0 ;
	}
}

int free_mem( ps, ou, nu, ov, nv, id, var )
	PARAMS ps ;
	double **ou, **nu, **ov, **nv, **id, **var ;
{
	free( ou ) ;
	free( nu ) ;
	free( ov ) ;
	free( nv ) ;
	free( id ) ;
	free( var ) ;

	return 0 ;
}

int all_mem( ps, pou, pnu, pov, pnv, pid, pvar )
	PARAMS ps ;
	double ***pou, ***pnu, ***pov, ***pnv, ***pid, ***pvar ;
{
	*pou = (double **)an_malloc_2D( ps.lx, ps.ly, sizeof ***pou ) ;
	*pnu = (double **)an_malloc_2D( ps.lx, ps.ly, sizeof ***pnu ) ;
	*pov = (double **)an_malloc_2D( ps.lx, ps.ly, sizeof ***pov ) ;
	*pnv = (double **)an_malloc_2D( ps.lx, ps.ly, sizeof ***pnv ) ;
	*pid = (double **)an_malloc_2D( ps.lx, ps.ly, sizeof ***pid ) ;
	*pvar = (double **)an_malloc_2D( ps.lx, ps.ly, sizeof ***pvar ) ;

	if ( *pou==NULL || *pnu==NULL ||
		*pov==NULL || *pnv==NULL || *pid==NULL || *pvar==NULL ) {
		free_mem( ps, *pou, *pnu, *pov, *pnv, *pid, *pvar ) ;
		return -1 ;
	} else
		return 0 ;
}

int main( argc, argv )
	int argc ;
	char **argv ;
{
	PARAMS ps ;
	double **ou, **nu, **ov, **nv, **id, **var ;
	int calc() ;

	if ( argc!=3 ) {
		fprintf( stderr, "usage: %s file.dat file\n", COMMAND_NAME ) ;
		return -1 ;
	} else if ( chk_para( argv, &ps )==-1 ) {
		fputs( "error: check the parameter file", stderr ) ;
		return -2 ;
	} else if ( all_mem( ps, &ou, &nu, &ov, &nv, &id, &var )==-1 ) {
		fputs( "error: cannot allocate memory\n", stderr ) ;
		return -2 ;
	} else if ( calc( ps, ou, nu, ov, nv, id, var )==-1 ) {
		free_mem( ps, ou, nu, ov, nv, id, var ) ;
		fputs( "error: cannot calculate", stderr ) ;
		return -2 ;
	} else {
		free_mem( ps, ou, nu, ov, nv, id, var ) ;
		return 0 ;
	}
}

